/******************************************************************************
*	Module: ds1302.c
*	Device Driver for Dallas DS1302 Real-Time-Clock (Timekeeper)
*
* 	Copyright (C) 2004 Volker Troyke, www.troyke.de
*
* 	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

#include <avr/interrupt.h>
#include "global.h"
#include "DS1302.h"

// 06.12.2003 	-	Timezone stored in DS1302 RAM location 0, Changed time structure -> timezone byte added
// 11.01.2004	-	write protect now only disabled in write routines, corrected initialization, bit 7 of seconds reg now always written to 0 -> clock enable
// 17.01.2004	-	bugfix while writing to ds1302, day wasn't bcd-coded before writing

u08 bin2bcd(u08 bin)
{
	u08 zehner, rest;

	zehner = 0;
	rest = bin;

	while (rest >= 10)
	{
		zehner++;
		rest = rest - 10;
	}

	return ((zehner << 4) | rest);
}

u08 bcd2bin(u08 bcd)
{
	u08 bin, i;

	bin=0;
	bin=(bcd & 0x0f);
	for(i=0; i< (bcd>>4); i++)
	{
		bin += 10;
	}
	return bin;
}

void write_command_byte_ds1302(u08 command)
{
	u08 i;

	sbi(DDR_1302,IO);
	sbi(PORT_1302,RST_L);

	for(i=0;i<10;i++);

	for(i=0;i<8;i++)
	{
		if((command>>i) & 0x01)
		{
			sbi(PORT_1302,IO);
		}
		else
		{
			cbi(PORT_1302,IO);
		}

		sbi(PORT_1302,SCLK);
		if(i<7)
		{
			cbi(PORT_1302,SCLK);
		}
	}
}

void write_data_byte_ds1302(u08 data)
{
	u08 i;

	cbi(PORT_1302,SCLK);
	for(i=0;i<8;i++)
	{
		if((data>>i) & 0x01)
		{
			sbi(PORT_1302,IO);
		}
		else
		{
			cbi(PORT_1302,IO);
		}

		sbi(PORT_1302,SCLK);
		cbi(PORT_1302,SCLK);
	}
}

u08 read_data_byte_ds1302(void)
{
	u08 i, temp;

	cbi(DDR_1302,IO); 	// IO Input
	cbi(PORT_1302,IO); 	// Pullup off
	cbi(PORT_1302,SCLK); // Clock low

	temp=0;

	for(i=0;i<8;i++)
	{
		if(bit_is_set(PIN_1302, IO))
		{
			temp |= (1<<i);
		}
		sbi(PORT_1302,SCLK); // Clock high
		cbi(PORT_1302,SCLK); // Clock low
	}
	return temp;
}

void init_ds1302(void)
{
// set 1302 ports as output
	sbi(DDR_1302, SCLK);
	sbi(DDR_1302, RST_L);
	sbi(DDR_1302, IO);

// set 1302 port bits to default values
	cbi(PORT_1302, SCLK);
	cbi(PORT_1302, RST_L);
	cbi(PORT_1302, IO);

	write_byte_ds1302(0x00, 0x07);	// write protect off
	write_byte_ds1302(TRICKLE_CHARGE_VALUE, 0x08);	// trickle charge setting
	write_byte_ds1302(0x80, 0x07);	// write protect on
}

void write_byte_ds1302(u08 data, u08 adr)
{
	write_command_byte_ds1302(0x80 | (adr<<1));
	write_data_byte_ds1302(data);
	cbi(PORT_1302,RST_L);
}

u08 read_byte_ds1302(u08 adr)
{
	u08 temp;
	write_command_byte_ds1302(0x81 | (adr<<1));
	temp=read_data_byte_ds1302();
	cbi(PORT_1302,RST_L);
	return temp;
}

void read_time_ds1302(struct dt *date_time)
{
	u08 i, temp;

	write_command_byte_ds1302(0xBF);

	for(i=0;i<7;i++)
	{
		temp=read_data_byte_ds1302();
		temp = bcd2bin(temp);
		switch(i)
		{
			case 0:
				date_time->t.s=temp;
				break;
			case 1:
				date_time->t.m=temp;
				break;
			case 2:
				date_time->t.h=temp;
				break;
			case 3:
				date_time->d.tag=temp;
				break;
			case 4:
				date_time->d.monat=temp;
				break;
			case 5:
				date_time->d.wt=temp;
				break;
			case 6:
				date_time->d.jahr=temp;
				break;
		}
	}
	cbi(PORT_1302,RST_L);
	temp=read_byte_ds1302(0x20); // RAM Adress 0
	date_time->t.timezone = temp;
}

void write_time_ds1302(struct dt *date_time)
{
	u08 i, temp;

	write_byte_ds1302(0x00, 0x07);	// write protect off

	write_command_byte_ds1302(0xBE);

	for(i=0;i<8;i++)
	{
		switch(i)
		{
			case 0:
				temp = (bin2bcd(date_time->t.s)) & 0x7F; // always enable clock oszillator
				break;
			case 1:
				temp = bin2bcd(date_time->t.m);
				break;
			case 2:
				temp =(bin2bcd(date_time->t.h)) & 0x7F; // set always 24 hour mode
				break;
			case 3:
				temp = bin2bcd(date_time->d.tag);
				break;
			case 4:
				temp = bin2bcd(date_time->d.monat);
				break;
			case 5:
				temp = bin2bcd(date_time->d.wt);
				break;
			case 6:
				temp = bin2bcd(date_time->d.jahr);
				break;
			case 7:
				temp = 0x00;
				break;
			default:
				temp = 0x00;
		}
		write_data_byte_ds1302(temp);
	}
	cbi(PORT_1302,RST_L);
	write_byte_ds1302(date_time->t.timezone,0x20); // RAM Adress 0

	write_byte_ds1302(0x80, 0x07);	// write protect on
}
