/******************************************************************************
*	Module: ntp_server.c
*	NTP-Server synchronized to DCF77 Radio Clock (Germany Mainflingen 77,5kHz)
*
*	Funtion ethernet_process_data based on:
*  The uIP TCP/IP stack code author Adam Dunkels <adam@dunkels.com>
*
* 	Copyright (C) 2004 Volker Troyke, www.troyke.de
*
* 	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/


// History
// Version 1.1 - 02.05.2006 uip_arp.c: ARP-Table Timeout fixed, ntp_server.c: priodically ARP-Request to Dafault Gate added
// Version 1.2 - 24.07.2006 dcf77.h: toleranz values extended MAX_HIGH changed from 230 to 245, MIN_HIGH changed from 180 to 170
//									 ntp_server.c: Indication of dcf signal fail changed
// Version 1.3 - 01.03.2008 ntp_server.c: Fehler in der Schaltjahrkorrektur im laufenden Jahr
// Version 1.4 - 17.04.2008 ntp_server.c, telnet.c: Display Telnet Login IP
// Version 1.5 - 01.02.2012 new Compiler WINAVR 20010110, lcd.c volatile Bugfix delay-function

#include <avr\interrupt.h>
#include "global.h"
#include "uip.h"
#include "nic.h"
#include "uip_arp.h"
#include "lcd.h"
#include "dcf77.h"
#include "ds1302.h"
#include "telnet.h"

#include "compiler.h"


#define BUF ((struct uip_eth_hdr *)&uip_buf[0])
#define IPBUF ((uip_tcpip_hdr *)&uip_buf[UIP_LLH_LEN])

struct ntp_frame
{
	u08 flag;
	u08 peer_clock_stratum;
	u08 peer_polling_intervall;
	u08 peer_clock_precision;
	u32 root_delay;
	u32 clock_dispersion;
	u08 reverence_clock_id[4];
	u32 reference_clock_update_i;
	u32 reference_clock_update_f;
	u32 originate_timestamp_i;
	u32 originate_timestamp_f;
	u32 receive_timestamp_i;
	u32 receive_timestamp_f;
	u32 transmit_timestamp_i;
	u32 transmit_timestamp_f;
	u16 debug;
};

#if TICK_100M > 255
u16 prescale;
#else
u08 prescale;
#endif

#if TICK_UIP > 255
u16 tick_uip;
#else
u08 tick_uip;
#endif

struct _ipconfig ipconfig;
struct _database database;

struct _baud baud[MAXBAUD] = {
	{"1200", 0x03, 0x54},
	{"2400", 0x01, 0xA9},
	{"4800", 0x00, 0xD4},
	{"9600", 0x00, 0x69},
	{"19.2", 0x00, 0x34},
	{"38.4", 0x00, 0x1A},
	{"57.6", 0x00, 0x11},
	{"115.", 0x00, 0x08}
};

u08 volatile update, uip_periodic_timeout;
u08 display_ntp;

u16 ntp_req_counter;

char data[6];
struct dt act_date;
struct dt dcf_date;
u32 update_timestamp;
u32 ntp_time_i;
u08 dcf_update_done;
u16 arptimer;
u08 alive;

u08 rx_buffer[256];
u08 tx_buffer[256];
u08 rx_writepointer;
u08 rx_readpointer;
u08 tx_writepointer;
u08 tx_readpointer;

u08 uart_tx;
u08 connect;

u16 uart_connect_ip[2];
u16 telnet_connect_ip[2];
u16 tx_bytes, rx_bytes;
u16 rx_overflow, tx_overflow;

u16 ntp_last_connect_ip[2];

u16 uart_timeout;
u16 telnet_timeout;

void initTimer(void)
{
	prescale = 0;
	update = 0;
	uip_periodic_timeout = 0;
	tick_uip = 0;

 	TCCR0 = 0x0B;	   // timer0 prescale 1/64, clear on Compare match
	OCR0	= 124;		// compare match on 124 (counting 123..124..0..1) every 125 count one INT

  	// interrupt on compare match (TOP of Counter, set in OCR0)
  	sbi(TIMSK, OCIE0) ;
}

SIGNAL(SIG_OUTPUT_COMPARE0)
{
  	dcf77_timetick();

	if(prescale == TICK_100M - 1)	// 100ms tick
	{
		prescale = 0;
		update = 1;
	}
	else
	{
		prescale++;
	}

	if(tick_uip == (TICK_UIP - 1))	// UIP periodic Timer
//	if(tick_uip == (pgm_read_byte(&baud[database.baud_rate].poll_rate) - 1))	// UIP periodic Timer
	{
		tick_uip = 0;
		uip_periodic_timeout = 1;
	}
	else
	{
		tick_uip++;
	}
}


void ethernet_process_data(void)
{
	u08 i;
	// look for a packet
	uip_len = nic_poll();
	if(uip_len == 0)
	{
	   // if timed out, call periodic function for each connection
	   if(uip_periodic_timeout)
	   {
	   	uip_periodic_timeout = 0;

	     	for(i = 0; i < UIP_CONNS; i++)
	     	{
	       	uip_periodic(i);

	       	// transmit a packet, if one is ready
	       	if(uip_len > 0)
	       	{
	         	uip_arp_out();
	         	nic_send();
	         	uip_auto_ack();
	       	}
	     	}

	     	for(i = 0; i < UIP_UDP_CONNS; i++)
	     	{
	       	uip_udp_periodic(i);

	       	// transmit a packet, if one is ready
	       	if(uip_len > 0)
	       	{
	         	uip_arp_out();
	         	nic_send();
	       	}
	     	}

	     	/* Call the ARP timer function every 10 seconds. */
	     	if(++arptimer == 1000)
	     	{
	       	uip_arp_timer();
	       	arptimer = 0;

	       	IPBUF->destipaddr[0] = ipconfig.gateway[0]; // poll default gate for MAC
	       	IPBUF->destipaddr[1] = ipconfig.gateway[1];
	       	uip_arp_out();
	     		if(uip_len > 0)
	       	{
		     		nic_send();
		     	}
	     	}
	   }
	}
	else  // packet received
	{
		// process an IP packet
	   if(BUF->type == htons(UIP_ETHTYPE_IP))
	   {
	     	// add the source to the ARP cache
	     	// also correctly set the ethernet packet length before processing
	     	uip_arp_ipin();
	     	uip_input();

	     	// transmit a packet, if one is ready
	     	if(uip_len > 0)
	     	{
	       	uip_arp_out();
	       	nic_send();
	       	uip_auto_ack();
	    	}
	   }
	   // process an ARP packet
	   else if(BUF->type == htons(UIP_ETHTYPE_ARP))
	   {
	     	uip_arp_arpin();

	     	// transmit a packet, if one is ready
	     	if(uip_len > 0)
	     	{
	      	nic_send();
	      }
	   }
	}
}

void display_date(struct dt *date)
{
	char data[5];

  	lcd_gotoxy(1,1);

	switch(date->d.wt)
	{
		case MO: lcd_puts_P(PSTR("Mo")); break;
		case DI: lcd_puts_P(PSTR("Di")); break;
		case MI: lcd_puts_P(PSTR("Mi")); break;
		case DO: lcd_puts_P(PSTR("Do")); break;
		case FR: lcd_puts_P(PSTR("Fr")); break;
		case SA: lcd_puts_P(PSTR("Sa")); break;
		case SO: lcd_puts_P(PSTR("So")); break;
	}

	lcd_puts_P(PSTR(" "));

	if(date->d.tag < 10)
	{
		lcd_puts_P(PSTR(" "));
		char2str(date->d.tag, data,1);
		lcd_puts(data);
	}
	else
	{
		char2str(date->d.tag, data,2);
		lcd_puts(data);
	}

	lcd_puts_P(PSTR("."));
	char2str(date->d.monat, data,2);
	lcd_puts(data);
	lcd_puts_P(PSTR("."));
	char2str(date->d.jahr, data,2);
	lcd_puts(data);

	if(!signal_ok)
	{
		lcd_puts_P(PSTR("\002"));
	}
	else
	{
		if(!dcf_valid)
		{
			lcd_puts_P(PSTR("\003"));
		}
		else
		{
			lcd_puts_P(PSTR(" "));
		}
	}


	if(date->t.h < 10)
	{
		lcd_puts_P(PSTR(" "));
		char2str(date->t.h, data,1);
		lcd_puts(data);
	}
	else
	{
		char2str(date->t.h, data,2);
		lcd_puts(data);
	}

	lcd_puts_P(PSTR(":"));

	char2str(date->t.m, data,2);
	lcd_puts(data);

	lcd_puts_P(PSTR(":"));

	char2str(date->t.s, data,2);
	lcd_puts(data);
}

u32 calc_act_timestamp(struct dt *date)
{
	u32 temp;
	u08 i;

	temp=3155673600L; // Sekunden bis 1.1.2000


	temp += ((date->d.jahr) * 365L*3600L*24L);

	temp += (((date->d.jahr - 1) / 4L) * 3600L*24L); // Schaltjahre Korrektur

	i=1;
	while(i < date->d.monat)
	{
		temp += (pgm_read_byte(&tage_im_monat[i]) * 3600L*24L);
		i++;
	}

//	if( (date->d.monat > 3) && ( ((date->d.jahr & 0x03) == 0) ) )	// Fehler koriggiert 1.3.2008

	if( (date->d.monat > 2) && ( ((date->d.jahr & 0x03) == 0) ) ) 	// Schaltjahrkorrektur laufendes (Schalt)Jahr
	{
		temp += 3600L*24L;
	}

	temp += (date->d.tag * 3600L*24L);
	temp += (date->t.h *3600L);
	temp += (date->t.m *60L);
	temp += date->t.s;

	if(date->t.timezone & 0x01)
	{
		temp -= 3600L*2; 	// Korrektur UTC und Sommerzeit
	}
	else
	{
		temp -= 3600L; 	// Korrektur UTC
	}
	return temp;
}

void display_ip(u16_t *ip)
{
	char2str(ip[0] & 0x00FF, data,3);
	lcd_puts(data);
	lcd_puts_P(PSTR("."));

	char2str(ip[0] >> 8, data,3);
	lcd_puts(data);
	lcd_puts_P(PSTR("."));

	char2str(ip[1] & 0x00FF, data,3);
	lcd_puts(data);
	lcd_puts_P(PSTR("."));

	char2str(ip[1] >> 8, data,3);
	lcd_puts(data);
}

void send_rx_data(void)
{
	u16 i;

	i = 0;
	while((rx_writepointer != rx_readpointer) && (i < UIP_TCP_MSS) && (i < uip_mss()) )
	{
		uip_appdata[i] = rx_buffer[rx_readpointer];
		rx_readpointer++;
		i++;
	};

	if(i > 0)
	{
		uip_send(uip_appdata, i);
		noack();
	}
}

void reset_uart_timeout(void)
{
	uart_timeout=database.uart_timeout;
}

void reset_telnet_timeout(void)
{
	telnet_timeout=TELNET_TIMEOUT;
}

void tcp_appcall(void)
{
	u08 i;

	if(uip_newdata())
	{
		if(uip_conn->lport == HTONS(PORT_UART))
		{
			if(connect && uart_connect_ip[0] == uip_conn->ripaddr[0] && uart_connect_ip[1] == uip_conn->ripaddr[1])
			{
				i = 0;
				while(i < uip_len)
				{
					if(((tx_writepointer + 1) & 0x00FF) != tx_readpointer)
					{
						tx_buffer[tx_writepointer] = uip_appdata[i];
						tx_writepointer++;
						i++;
					}
					else
					{
						i = uip_len;  // abort, tx_buffer ist full
						tx_overflow++;
					}
				};

				if(((tx_writepointer - tx_readpointer) & 0x00FF) > (0xFF - UIP_RECEIVE_WINDOW)) // stop receiver if there ist not enought space left for a Packet of receive Window size
				{
					uip_stop();
				}
				reset_uart_timeout();
				send_rx_data();
			}
		}

		if(uip_conn->lport == HTONS(PORT_TELNET))
		{
			telnet();
			reset_telnet_timeout();
		}
	}

	if(uip_acked() && !uip_newdata())
	{
		if(uip_conn->lport == HTONS(PORT_UART))
		{
			if(connect && uart_connect_ip[0] == uip_conn->ripaddr[0] && uart_connect_ip[1] == uip_conn->ripaddr[1])
			{
				reset_uart_timeout();
			}
		}

		if(uip_conn->lport == HTONS(PORT_TELNET))
		{
			reset_telnet_timeout();
		}
	}

	if(uip_poll())
	{
		if(uip_conn->lport == HTONS(PORT_UART))
		{
			if(connect && uart_connect_ip[0] == uip_conn->ripaddr[0] && uart_connect_ip[1] == uip_conn->ripaddr[1])
			{
				if((((tx_writepointer - tx_readpointer) & 0x00FF) <= (0xFF - UIP_RECEIVE_WINDOW)) && uip_stopped())
				{
					uip_restart();
				}

				if(!uart_timeout && !uip_closed())
				{
					uip_close();
				}
				send_rx_data();
			}
		}

		if(uip_conn->lport == HTONS(PORT_TELNET))
		{
			if(!telnet_timeout)
			{
				tnpos = 0;
				ip_addstring_P(PSTR("."));
				ip_sendstring();
				reset_telnet_timeout();
			}
		}
	}

	if(uip_connected())
	{
		if(uip_conn->lport==HTONS(PORT_TELNET))
		{
			lcd_gotoxy(17,2);
			lcd_puts_P(PSTR("Teln"));
			tnpos=0;
			telnet_connect_ip[0] = uip_conn->ripaddr[0];
			telnet_connect_ip[1] = uip_conn->ripaddr[1];
			ip_addstring_P(PSTR("   -- UART/NTP-Server Telnet Console -- \r\n"));
			ip_addstring_P(PSTR("   (C) Volker Troyke 2012 www.troyke.de \r\n\r\n"));
			ip_addstring_P(PSTR("SW-Version 1.5 from: "__DATE__" "__TIME__"\r\n"));
#ifndef LCD_BUS_8
			ip_addstring_P(PSTR("LCD 4-Bit-Version\r\n"));
#endif
#ifndef LCD_TWO_LINES
			ip_addstring_P(PSTR("LCD 4-Line-Version\r\n"));
#endif
			ip_addstring_P(PSTR("\r\n"));
			ip_addstring_P(PSTR("Login from IP: "));
			display_ip_telnet(telnet_connect_ip);
			ip_addstring_P(PSTR("\r\n\r\n"));
			ip_addstring_P(PSTR("use h or ? for Help\r\n"));
			ip_sendstring();
			inputtype=INPUT_IDLE;
			lcd_gotoxy(1,2);
			display_ip(uip_conn->ripaddr);
			lcd_puts_P(PSTR(">"));
			display_ntp = DISPLAY_IP_TIME;
			reset_telnet_timeout();
		}
		else
		{
			if(connect==0)
			{
				UBRRH=pgm_read_byte(&baud[database.baud_rate].baud_reg_h);		// set baud rate H-Register
				UBRRL=pgm_read_byte(&baud[database.baud_rate].baud_reg_l);		// set baud rate L-Register

				lcd_gotoxy(17,2);
				lcd_puts_P(baud[database.baud_rate].text);

				rx_writepointer=0;
				rx_readpointer=0;
				tx_writepointer=0;
				tx_readpointer=0;
				rx_bytes=0;
				tx_bytes=0;
				uart_connect_ip[0]=uip_conn->ripaddr[0];
				uart_connect_ip[1]=uip_conn->ripaddr[1];
				cbi(PORT_RTS,RTS);
				uip_unlisten(HTONS(PORT_UART));
				connect=1;
				lcd_gotoxy(1,2);
				display_ip(uip_conn->ripaddr);
				lcd_puts_P(PSTR(">"));
				display_ntp = DISPLAY_IP_TIME;
				reset_uart_timeout();
			}
		}
	}

	if( uip_closed() || uip_aborted() || uip_timedout() )
	{
		if(uip_conn->lport == HTONS(PORT_UART))
		{
			if(connect && uart_connect_ip[0]==uip_conn->ripaddr[0] && uart_connect_ip[1]==uip_conn->ripaddr[1])
			{
				lcd_gotoxy(16,2);
				lcd_puts_P(PSTR(" disc"));
				sbi(PORT_RTS,RTS);
				connect=0;
				uart_timeout=0;
				uip_listen(HTONS(PORT_UART));
			}
		}

		if(uip_conn->lport == HTONS(PORT_TELNET))
		{
			if(!connect)
			{
				lcd_gotoxy(16,2);
				lcd_puts_P(PSTR(" disc"));
			}
			else
			{
				lcd_gotoxy(16,2);
				lcd_puts_P(PSTR(" "));
				lcd_puts_P(baud[database.baud_rate].text);
			}
		}
	}
}

void udp_appcall(void)
{
	u08 i;
	struct ntp_frame *ntp_dat;

	if(uip_flags == UIP_POLL)
	{
		uip_udp_conn->lport=0;
		uip_slen = 0;
	}
	else
	{
		if(uip_udp_conn->lport==HTONS(123))
		{
			lcd_gotoxy(1,2);
			display_ip(uip_udp_conn->ripaddr);
			ntp_last_connect_ip[0]=uip_udp_conn->ripaddr[0];
			ntp_last_connect_ip[1]=uip_udp_conn->ripaddr[1];
			lcd_puts_P(PSTR(" "));
			ntp_req_counter++;
			display_ntp = DISPLAY_IP_TIME;

			ntp_dat = (struct ntp_frame *)uip_appdata;

			ntp_dat->peer_clock_stratum			= 0x01; // primary Reference
		//	ntp_dat->peer_polling_intervall		= 8;

			ntp_dat->root_delay						= 0;
			ntp_dat->clock_dispersion				= 0;
			ntp_dat->reverence_clock_id[0]		= 'D';
			ntp_dat->reverence_clock_id[1]		= 'C';
			ntp_dat->reverence_clock_id[2]		= 'F';
			ntp_dat->reverence_clock_id[3]		= 0;
			for(i=0;i<4;i++)
			{
				((u08 *)&ntp_dat->reference_clock_update_i)[i]=((u08 *)&update_timestamp)[3-i];
			}
			ntp_dat->reference_clock_update_f	= 0;
			ntp_dat->originate_timestamp_i		= ntp_dat->transmit_timestamp_i;
			ntp_dat->originate_timestamp_f		= ntp_dat->transmit_timestamp_f;

			cli();
			for(i=0;i<4;i++)
			{
				((u08 *)&ntp_dat->transmit_timestamp_i)[i]=((u08 *)&ntp_time_i)[3-i];
			}
			((u08 *)&ntp_dat->transmit_timestamp_f)[0]			= dcf77_tick >> 3;
			((u08 *)&ntp_dat->transmit_timestamp_f)[1]			= dcf77_tick << 5;
			sei();

			if(dcf_valid || dcf_update_done)
			{
				ntp_dat->flag							= 0x1c;		// Normal Operation
				ntp_dat->peer_clock_precision		= -11;
				((u08 *)&ntp_dat->transmit_timestamp_f)[2] = 0;
				((u08 *)&ntp_dat->transmit_timestamp_f)[3] = 0;
			}
			else
			{
				ntp_dat->flag							= 0xdc;		// Alarm Condition not synchronized
				ntp_dat->peer_clock_precision		= 0;
				ntp_dat->transmit_timestamp_f 	= 0xFFFFFF7F;
			}

			ntp_dat->receive_timestamp_i			= ntp_dat->transmit_timestamp_i;
			ntp_dat->receive_timestamp_f			= ntp_dat->transmit_timestamp_f;

			uip_slen = 48; // normale Lnge
		}
	}
}

SIGNAL(SIG_UART_RECV)
{
	u08 data, status;

	status = UCSRA;				  // read status
	data = UDR;
	status &= 0x18;
	if(status)
	{
		// abort receive
	}
	else
	{
		if(connect)
		{
			if(((rx_writepointer + 1) & 0x00FF) == rx_readpointer)
			{
				rx_overflow++;
			}
			else
			{
				rx_buffer[rx_writepointer] = data;
				rx_writepointer++;
				rx_bytes++;
			}
		}
	}
}

// UART Data Register Empty Interrupt Function
SIGNAL(SIG_UART_DATA)
{
	if(tx_writepointer != tx_readpointer)
	{
		UDR=tx_buffer[tx_readpointer];
		tx_readpointer++;
		tx_bytes++;
	}
	else
	{
		cbi(UCSRB,UDRIE); // disable empty interrupt
		uart_tx=0;
	}
}


/*****************************************************************************
*  Main Control Loop
*
*
*****************************************************************************/
int main(void)
{
	u08 i;
	u08 data[6];

	ACSR=0x80;  						// analog comparator switched off

	arptimer = 0;
	alive = 0;
	ntp_req_counter = 0;
	update_timestamp = 0;
	dcf_update_done = 0;

	GICR =0x40;							// Enable Int 0 for DCF77 Receiver
	MCUCR=0x02;							// Int 0 falling Edge

	cbi(DDRD,PD2);						// switch DCF77 Port to input
	cbi(PORTD,PD2);					// high Impedance

	cbi(DDR_IPRESET, IPRESET);    // IP-Configuration Reset Pin to Input
	sbi(PORT_IPRESET, IPRESET);   // IP-Configuration Reset Pin enbale Pullup

	sbi(DDR_RTS,RTS);					// switch RTS Pin to out
	sbi(PORT_RTS,RTS);				// RTS disabled

  	initTimer();
	dcf77_init();
	init_ds1302();

// setup UART
	UCSRA=0x00;
	UCSRB=0x98;
	UCSRC=0x86;

	read_time_ds1302(&act_date);

	ntp_time_i = calc_act_timestamp(&act_date);

	for(i=0;i<4;i++)
	{
		((u08 *)&update_timestamp)[i] = read_byte_ds1302(0x25+i); // RAM Adress 5+i
	}

	eeprom_read_block(&ipconfig, (int*)IP_CONFIG_BASE, sizeof(struct _ipconfig));
	eeprom_read_block(&database, (int*)DATABASE_BASE,  sizeof(struct _database));

	if(database.baud_rate >= MAXBAUD)
	{
		database.baud_rate = 0;
	}

	if(ipconfig.hostadr[0] == 0xFFFF || !(PIN_IPRESET && _BV(IPRESET) ) )
	{
		ipconfig.hostadr[0] = 192 + (168 <<8);
		ipconfig.hostadr[1] = 100 + ( 99 <<8);
		ipconfig.netmask[0] = 255 + (255 <<8);
		ipconfig.netmask[1] = 255 + (  0 <<8);
		ipconfig.gateway[0] = 192 + (168 <<8);
		ipconfig.gateway[1] = 100 + (  1 <<8);
		eeprom_write_block(&ipconfig, (int*)IP_CONFIG_BASE, sizeof(struct _ipconfig));

		database.baud_rate = 0;
		database.uart_timeout = 1800; // 3 min in 100ms ticks
		eeprom_write_block(&database, (int*)DATABASE_BASE,  sizeof(struct _database));
	}

	sei();

	rx_writepointer=0;
	rx_readpointer=0;
	tx_writepointer=0;
	tx_readpointer=0;
	uart_tx=0;
	connect=0;
	tx_bytes=0;
	rx_bytes=0;
	rx_overflow = 0;
	tx_overflow = 0;
	ntp_last_connect_ip[0]=0;
	ntp_last_connect_ip[1]=0;

	// init NIC device driver
	nic_init();

	// init uIP
	uip_init();

	// init ARP cache
	uip_arp_init();

	uip_sethostaddr(ipconfig.hostadr);
	uip_setnetmask(ipconfig.netmask);
	uip_setdraddr(ipconfig.gateway);

	uip_listen(HTONS(PORT_UART));
	uip_listen(HTONS(PORT_TELNET));

	lcd_init();

	lcd_mode(LCD_ON_CURSOR_OFF);

	// SET Charactercode for ASCII-Code 0x02 (ICON )
	lcd_setchar(0x02, 0, 0x04);
	lcd_setchar(0x02, 1, 0x04);
	lcd_setchar(0x02, 2, 0x00);
	lcd_setchar(0x02, 3, 0x00);
	lcd_setchar(0x02, 4, 0x00);
	lcd_setchar(0x02, 5, 0x00);
	lcd_setchar(0x02, 6, 0x00);
	lcd_setchar(0x02, 7, 0x00);

	// SET Charactercode for ASCII-Code 0x03 (ICON )
	lcd_setchar(0x03, 0, 0x04);
	lcd_setchar(0x03, 1, 0x00);
	lcd_setchar(0x03, 2, 0x00);
	lcd_setchar(0x03, 3, 0x00);
	lcd_setchar(0x03, 4, 0x00);
	lcd_setchar(0x03, 5, 0x00);
	lcd_setchar(0x03, 6, 0x00);
	lcd_setchar(0x03, 7, 0x00);

	lcd_gotoxy(1,1);
	lcd_puts_P(PSTR("  NTP/UART-Server   "));

	lcd_gotoxy(1,2);
	display_ip(uip_hostaddr);
	lcd_puts_P(PSTR(" disc"));

	display_ntp = 120;

	wdt_enable(WDTO_250MS);

	while(1)
	{
		ethernet_process_data();

		if(!uart_tx)
		{
			if(tx_writepointer != tx_readpointer)
			{
				UDR=tx_buffer[tx_readpointer];
				tx_readpointer++;
				tx_bytes++;
				uart_tx=1;
				sbi(UCSRB,UDRIE);  // empty interrupt enable
			}
		}


		if(update)							// 100ms Call Intervall
		{
			update = 0;

			wdt_reset();

			if(telnet_timeout)
			{
				telnet_timeout--;
			}

			if(uart_timeout && (uart_timeout != 65535))
			{
				uart_timeout--;
			}
	   }

	   if(dcf77_update)					// new DCF-Bit received
		{
			dcf77_update = 0;
			dcf77_process_data();
		}

		if(update_time)					// DCF Time Information complete, update display Time and Timekeeper
		{
			update_time 			= 0;
			dcf_update_done		= 1;
			ntp_time_i 				= calc_act_timestamp(&dcf_date);
			update_timestamp 		= ntp_time_i;
			act_date.t.s 			= dcf_date.t.s;
			act_date.t.m 			= dcf_date.t.m;
			act_date.t.h 			= dcf_date.t.h;
			act_date.t.timezone 	= dcf_date.t.timezone;
			act_date.d.tag 		= dcf_date.d.tag;
			act_date.d.wt 		   = dcf_date.d.wt;
			act_date.d.monat 	   = dcf_date.d.monat;
			act_date.d.jahr 	   = dcf_date.d.jahr;

			write_time_ds1302(&act_date);

			write_byte_ds1302(0x00, 0x07);	// write protect off
			for(i=0;i<4;i++)
			{
				write_byte_ds1302(((u08 *)&update_timestamp)[i],0x25+i); // RAM Adress 5+i
			}
			write_byte_ds1302(0x80, 0x07);	// write protect on

			display_date(&act_date);
		}

	   if(inc_time) 	// 1s Call Intervall
	   {
	   	inc_time = 0;
	   	increment_time(&act_date);
	   	display_date(&act_date);

	   	if(!display_ntp)
	   	{
		   	lcd_gotoxy(1,2);
				lcd_puts_P(PSTR("NTP-REQ: "));
				char2str(ntp_req_counter, data,5);
				lcd_puts(data);
				lcd_puts_P(PSTR("  "));
			}
			else
			{
				display_ntp--;
			}
	   }
	}
}
